//! Module containg the struct for storing 3D model

pub mod face;
pub mod material;

use std::collections::HashMap;
use std::collections::hash_map::Entry;

use glium::VertexBuffer;

use parser::{parse, ParserError};

use model::face::{FaceVertex, Face};
use model::material::Material;

use math::vector::{Vector2, Vector3};
use math::bounding_box::BoundingBox3;

use renderer::Renderer;

#[derive(Copy, Clone, Debug, PartialEq)]
/// A raw vertex, with its 3D coordinates, texture coordinates and normals.
pub struct Vertex {
    vertex:     [f64; 3],
    tex_coords: [f64; 2],
    normal:     [f64; 3],
}

implement_vertex!(Vertex, vertex, tex_coords, normal);

/// A part of a 3D model.
///
/// It contains a list of faces that correspond to the same material.
pub struct Part {

    /// Name of the corresponding material.
    pub material_name: Option<String>,

    /// List of all the faces in the part.
    pub faces: Vec<Face>,

    /// VertexBuffer for rendering.
    vertex_buffer: Option<VertexBuffer<Vertex>>,

}

impl Part {

    /// Creates a empty part from its material name.
    pub fn new(material_name: Option<String>) -> Part {
        Part {
            material_name: material_name,
            faces: vec![],
            vertex_buffer: None,
        }
    }

    /// Adds a face to the part.
    ///
    /// The face material will be changed according to the part material.
    pub fn add_face(&mut self, f: Face) {
        let mut f = f;
        f.material_name = self.material_name.clone();
        self.faces.push(f);
    }

    /// Returns a reference to the vertex buffer.
    pub fn vertex_buffer(&self) -> &Option<VertexBuffer<Vertex>> {
        &self.vertex_buffer
    }


}

/// A 3D model.
///
/// It contains all the data needed to do a rendering.
pub struct Model {

    /// List of all vertices.
    pub vertices: Vec<Vector3<f64>>,

    /// List of all texture coordinates.
    pub texture_coordinates: Vec<Vector2<f64>>,

    /// List of all normals.
    pub normals: Vec<Vector3<f64>>,

    /// Map associating the name of a material to the real material.
    pub materials: HashMap<String, Material>,

    /// The list of parts of the model.
    pub parts: Vec<Part>,

    /// The list of faces, in the order of the file
    pub faces: Vec<Face>,

    /// The part that is currently selected.
    current_part_index: Option<usize>,
}

/// Error when adding a face to a model.
pub enum ModelError {

    /// Index (first usize) received but greater than size (second usize).
    IndexError(usize, usize),

}

impl Model {

    /// Creates an empty model.
    pub fn new() -> Model {
        Model {
            vertices: vec![],
            texture_coordinates: vec![],
            normals: vec![],
            materials: HashMap::new(),
            parts: vec![],
            faces: vec![],
            current_part_index: None,
        }
    }

    /// Merges the model passed as parameter into the current model.
    pub fn merge(&mut self, other: Model) {
        let mut other = other;

        let vertex_offset = self.vertices.len();
        let tex_coord_offset = self.texture_coordinates.len();
        let normal_offset = self.normals.len();
        let face_offset = self.faces.len();

        self.vertices.append(&mut other.vertices);
        self.texture_coordinates.append(&mut other.texture_coordinates);
        self.normals.append(&mut other.normals);

        for part in &mut other.parts {
            let mut new_part = Part::new(part.material_name.clone());
            for face in &mut part.faces {
                for fv in &mut [&mut face.a, &mut face.b, &mut face.c] {
                    fv.vertex += vertex_offset;
                    if let Some(tex_coord) = fv.texture_coordinate {
                        fv.texture_coordinate = Some(tex_coord + tex_coord_offset);
                    }
                    if let Some(normal) = fv.normal {
                        fv.normal = Some(normal + normal_offset);
                    }
                }
                if let Some(id) = face.id {
                    face.id = Some(id + face_offset);
                }
                new_part.faces.push(face.clone());
                self.faces.push(face.clone());
            }
            self.parts.push(new_part);
        }

        for (name, material) in other.materials {
            let entry = self.materials.entry(name);
            if let Entry::Occupied(_) = entry {
                // Return error
            } else {
                entry.or_insert(material);
            }
        }
    }

    /// Creates a model from a file.
    pub fn from(path: &str) -> Result<Model, ParserError> {
        parse(path)
    }

    /// Computes the area of a 3D face of the model.
    pub fn face_area(&self, f: &Face) -> f64 {
        self.face_vertex_area(&f.a, &f.b, &f.c)
    }

    /// Computes the 3D area of a face from 3 face vertices.
    pub fn face_vertex_area(&self, a: &FaceVertex, b: &FaceVertex, c: &FaceVertex) -> f64 {
        let v1 = self.vertices[a.vertex];
        let v2 = self.vertices[b.vertex];
        let v3 = self.vertices[c.vertex];
        Vector3::area(v1, v2, v3)
    }

    /// Returns the name of the current material, i.e. the material corresponding to the current
    /// part.
    pub fn current_material_name(&self) -> Option<String> {
        if let Some(p) = self.current_part() {
            p.material_name.clone()
        } else {
            None
        }
    }

    /// Changes the part in order to select a specific material.
    ///
    /// If no part with such material exists, it will be created.
    pub fn change_part(&mut self, material_name: Option<String>) {
        self.current_part_index = None;
        for (index, part) in self.parts.iter().enumerate() {
            if part.material_name == material_name {
                self.current_part_index = Some(index);
            }
        }

        if self.current_part_index.is_none() {
            // Create a new part
            self.create_part(material_name.clone());
            self.current_part_index = Some(self.parts.len() - 1);
        }
    }

    /// Creates a part with a specific material.
    pub fn create_part(&mut self, material_name: Option<String>) {
        self.parts.push(Part::new(material_name));
    }

    /// Returns the current selected part.
    ///
    /// Returns None if the model contains no part.
    pub fn current_part(&self) -> Option<&Part> {
        if let Some(index) = self.current_part_index {
            Some(&self.parts[index])
        } else {
            None
        }
    }

    /// Returns the current selected part.
    ///
    /// Returns None if the model contains no part.
    pub fn current_part_mut(&mut self) -> Option<&mut Part> {
        if let Some(index) = self.current_part_index {
            Some(&mut self.parts[index])
        } else {
            None
        }
    }

    /// Adds a face to the current part of the model.
    ///
    /// Returns a ModelError if the face has out of bounds face vertices.
    pub fn add_face(&mut self, f: Face) -> Result<(), ModelError> {
        // Check that the indices are correct
        for face_vertex in [&f.a, &f.b, &f.c].iter() {
            if face_vertex.vertex >= self.vertices.len() {
                return Err(ModelError::IndexError(face_vertex.vertex, self.vertices.len()));
            }

            if let Some(tc) = face_vertex.texture_coordinate {
                if tc >= self.texture_coordinates.len() {
                    return Err(ModelError::IndexError(tc, self.texture_coordinates.len()));
                }
            }

            if let Some(n) = face_vertex.normal {
                if n >= self.normals.len() {
                    return Err(ModelError::IndexError(n, self.normals.len()));
                }
            }
        }

        let material_name = self.current_material_name();
        self.change_part(material_name.clone());

        // Add the id and the material name
        let mut f = f;
        f.id = Some(self.faces.len());
        f.material_name = material_name;

        self.current_part_mut().unwrap().add_face(f.clone());
        self.faces.push(f);

        Ok(())
    }

    /// Computes the bounding box of the model.
    ///
    /// This function works on the vertices. If there is an unused vertex, it will still count in
    /// the bounding box.
    pub fn bounding_box(&self) -> BoundingBox3<f64> {
        let mut bbox = BoundingBox3::new(self.vertices[0].clone(), self.vertices[0].clone());

        for vertex in &self.vertices {
            bbox.add_point(&vertex);
        }

        bbox

    }

    /// Computes stats on the area of the faces
    ///
    /// Returns a pair containing the average and the standard deviation.
    pub fn get_area_stats(&self) -> (f64, f64) {

        let mut areas = vec![];
        let mut avg = 0.0;

        for (index, face) in self.faces.iter().enumerate() {

            let area = Vector3::area(
                self.vertices[face.a.vertex],
                self.vertices[face.b.vertex],
                self.vertices[face.c.vertex],
            );

            areas.push((index, area));
            avg += area;

        }

        avg /= areas.len() as f64;

        let mut sd = 0.0;

        for &(_, area) in &areas {
            sd += (area - avg) * (area - avg);
        }

        sd = (sd / areas.len() as f64).sqrt();

        (avg, sd)

    }

    /// Centers and scales the model from a bounding box.
    pub fn center_and_scale_from_box(&mut self, bbox: &BoundingBox3<f64>) {
        // let bbox = self.bounding_box();
        let center = (bbox.min() + bbox.max()) / 2.0;
        let size = (bbox.max() - bbox.min()).norm();

        for vertex in &mut self.vertices {
            *vertex -= center;
            *vertex /= size;
        }
    }

    /// Centers and scales the model from its own bounding box.
    pub fn center_and_scale(&mut self) {
        let bbox = self.bounding_box();
        self.center_and_scale_from_box(&bbox);
    }

    /// Creates vertex buffers for each part of the model.
    pub fn build_vertex_buffers(&mut self, renderer: &Renderer) {
        for part in &mut self.parts {
            Model::build_vertex_buffer_for_part(
                &self.vertices,
                &self.texture_coordinates,
                &self.normals,
                part,
                renderer);
        }
    }

    /// Creates a vertex buffer that can be rendered for a part of the model.
    pub fn build_vertex_buffer_for_part(
        vertices: &Vec<Vector3<f64>>,
        texture_coordinates: &Vec<Vector2<f64>>,
        normals: &Vec<Vector3<f64>>,
        part: &mut Part,
        renderer: &Renderer) {

        let mut vertex_buffer = vec![];
        for face in &part.faces {
            for &&v in &[&face.a, &face.b, &face.c] {
                let vertex = vertices[v.vertex].into();
                let tex_coord = if let Some(tex_index) = v.texture_coordinate {
                    texture_coordinates[tex_index].into()
                } else {
                    [0.0, 0.0]
                };

                let normal = if let Some(normal_index) = v.normal {
                    normals[normal_index].into()
                } else {
                    [0.0, 0.0, 0.0]
                };

                vertex_buffer.push(Vertex {
                    vertex:     vertex,
                    tex_coords: tex_coord,
                    normal:     normal,
                });
            }
        }

        part.vertex_buffer = Some(renderer.build_vertex_buffer(&vertex_buffer));
    }

    /// Builds the textures of all materials.
    pub fn build_textures(&mut self, renderer: &Renderer) {
        for (_, ref mut material) in &mut self.materials {
            Model::build_texture_for_material(material, renderer);
        }
    }

    /// Builds the SrgbTextures for rendering.
    pub fn build_texture_for_material(material: &mut Material, renderer: &Renderer) {
        if let Some(path) = material.textures.get("map_Kd") {
            material.rendering_texture = Some(renderer.make_texture(path));
        };
    }
}
